import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/dialog/datetime_picker_dialog.dart';
import 'package:clean_android_design/src/utils/validation/date_validation.dart';
import 'package:flutter/material.dart';
import 'package:intl/intl.dart';
import 'package:provider/provider.dart';

import 'default_field_widget.dart';

class DatetimeFieldWidget<T> extends StatefulWidget implements IListItem {
  final bool isFirst;
  final bool isLast;
  final Text title;
  final bool withDate;
  final bool withTime;

  final DateTimeValidation<T> validation;
  final void Function(T, DateTime) saveToItem;
  final DateTime Function(T) getFromItem;

  final bool bound;
  final bool isEditable;

  const DatetimeFieldWidget(
      {Key key,
      this.title,
      this.saveToItem,
      this.getFromItem,
      this.withDate = true,
      this.withTime = true,
      this.validation})
      : assert(title != null),
        assert(getFromItem != null),
        assert(saveToItem != null),
        bound = true,
        isEditable = null,
        isLast = null,
        isFirst = null,
        super(key: key);

  const DatetimeFieldWidget.unbound({
    Key key,
    this.title,
    this.saveToItem,
    this.getFromItem,
    this.withDate = true,
    this.withTime = true,
    this.isEditable = true,
    this.validation,
  })  : assert(title != null),
        assert(getFromItem != null),
        assert(!isEditable || saveToItem != null),
        bound = false,
        isLast = null,
        isFirst = null,
        super(key: key);

  const DatetimeFieldWidget._copy({
    Key key,
    this.isFirst,
    this.isLast,
    this.title,
    this.withDate,
    this.withTime,
    this.saveToItem,
    this.getFromItem,
    this.bound,
    this.isEditable,
    this.validation,
  }) : super(key: key);

  @override
  _DatetimeFieldWidgetState<T> createState() => _DatetimeFieldWidgetState<T>();

  @override
  IListItem copyWithPosition({bool isFirst, bool isLast}) {
    return DatetimeFieldWidget<T>._copy(
      key: this.key,
      title: this.title,
      getFromItem: this.getFromItem,
      saveToItem: this.saveToItem,
      withDate: this.withDate,
      withTime: this.withTime,
      bound: this.bound,
      isEditable: this.isEditable,
      validation: this.validation,
      isFirst: isFirst ?? this.isFirst,
      isLast: isLast ?? this.isLast,
    );
  }
}

class _DatetimeFieldWidgetState<T> extends State<DatetimeFieldWidget<T>> {
  T item;
  String errorMessage;

  @override
  void initState() {
    super.initState();

    widget.validation?.initialize(
        getItemFunction: () => widget.bound ? Provider.of<T>(context) : null,
        getDateTimeFunction: () => widget.getFromItem(item),
        onValidate: (errorMessage) =>
            setState(() => this.errorMessage = errorMessage));
  }

  @override
  Widget build(BuildContext context) {
    if (widget.bound) item = Provider.of<T>(context);

    if (widget.isEditable == null)
      return Consumer<bool>(
          builder: (context, isEditable, __) => _getMainWidget(isEditable));
    else
      return _getMainWidget(widget.isEditable);
  }

  Widget _getMainWidget(bool isEditable) {
    return DefaultFieldWidget(
      isFirst: widget.isFirst,
      isLast: widget.isLast,
      separationLine: true,
      title: widget.title,
      childPadding:
          isEditable ? EdgeInsets.only(left: 30, right: 30, bottom: 3) : null,
      child: Column(
        children: <Widget>[
          Row(
            children: <Widget>[
              _getDateWidget(context, item, isEditable),
              SizedBox(
                width: widget.withDate ? 15 : 0,
              ),
              _getTimeWidget(item, isEditable),
            ],
          ),
          errorMessage != null
              ? Align(
                  alignment: Alignment.centerLeft,
                  child: Padding(
                    padding: const EdgeInsets.only(top: 6),
                    child: Text(
                      errorMessage,
                      style: TextStyles.textStyleSize(12).copyWith(
                          color: CleanAndroidDesignDefaultColors.error),
                    ),
                  ),
                )
              : Container(),
        ],
      ),
    );
  }

  Widget _getDateWidget(BuildContext context, T item, bool isEditable) {
    if (!widget.withDate) return Container();

    var date = widget.getFromItem(item);
    var dateString = date == null
        ? (isEditable ? "Datum" : "-")
        : DateFormat("EEE dd.MM.yyy", "de_DE").format(date);
    return isEditable
        ? SimpleButtonWidget(
            borderColor: errorMessage != null
                ? CleanAndroidDesignDefaultColors.error
                : null,
            width: 145,
            alignment: MainAxisAlignment.start,
            child: Text(
              dateString,
              style: TextStyles.textStyleSize(16).copyWith(
                color: CleanAndroidDesignDefaultColors.black,
              ),
            ),
            onPressed: () {
              FocusScope.of(context).unfocus();
              showCustomDialog(
                  context,
                  DatetimePickerDialog(
                    datetime: date,
                    setDatetime: (date) =>
                        setState(() => widget.saveToItem(item, date)),
                  ));
            },
          )
        : Text(
            dateString,
            style: TextStyles.textStyleSize(17)
                .copyWith(color: CleanAndroidDesignDefaultColors.black),
          );
  }

  Widget _getTimeWidget(T item, bool isEditable) {
    if (!widget.withTime) return Container();

    var date = widget.getFromItem(item);
    var dateString = date == null
        ? (isEditable ? "Uhrzeit" : (widget.withDate ? "" : "-"))
        : DateFormat("HH:mm").format(date);
    return isEditable
        ? SimpleButtonWidget(
            borderColor: errorMessage != null
                ? CleanAndroidDesignDefaultColors.error
                : null,
            width: date == null ? 90 : 70,
            alignment: MainAxisAlignment.start,
            child: Text(
              dateString,
              style: TextStyles.textStyleSize(16)
                  .copyWith(color: CleanAndroidDesignDefaultColors.black),
            ),
            onPressed: () {
              showCustomDialog(
                  context,
                  DatetimePickerDialog(
                    datetime: date,
                    timePicker: true,
                    setDatetime: (date) =>
                        setState(() => widget.saveToItem(item, date)),
                  ));
            },
          )
        : Text(
            dateString,
            style: TextStyles.textStyleSize(17)
                .copyWith(color: CleanAndroidDesignDefaultColors.black),
          );
  }
}
