import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:clean_android_design/src/dialog/save_or_abort_dialog.dart';
import 'package:clean_android_design/src/services/database_service_interface.dart';
import 'package:clean_android_design/src/services/models_interfaces.dart';
import 'package:clean_android_design/src/structure/simple_scaffold_menu_button_widget.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';
import 'package:provider/provider.dart';

import 'default_scaffold_widget.dart';
import 'icon_scaffold_menu_button_widget.dart';
import 'scaffold_page_widget.dart';

class DefaultUpdateScaffoldWidget<T extends UpdateableDatabaseItem<T>>
    extends StatefulWidget {
  static DatabaseServiceInterface databaseService;

  final T updateableItem;
  final Validator validator;
  final void Function(BuildContext context, T item) beforeSave;
  final Future<void> Function(BuildContext context, T item) beforeSaveAsync;
  final void Function(BuildContext context, DocumentReference) afterSave;
  final void Function(BuildContext context, T modifiedItem, T originalItem)
      onAbort;
  final void Function(BuildContext context, T item) prepareNewItem;

  final List<Widget> menuButtons;
  final List<Widget> editMenuButtons;
  final bool hasEditButton;

  final Widget Function(T modifiedItem) titleBuilder;
  final Widget child;
  final List<ScaffoldPageWidget<T>> Function(BuildContext context)
      updatePagesBuilder;

  const DefaultUpdateScaffoldWidget({
    Key key,
    this.updateableItem,
    this.menuButtons,
    this.hasEditButton = true,
    this.updatePagesBuilder,
    this.editMenuButtons,
    this.titleBuilder,
    this.validator,
    this.child,
    this.beforeSave,
    this.prepareNewItem,
    this.onAbort,
    this.beforeSaveAsync,
    this.afterSave,
  })  : assert(titleBuilder != null),
        assert(beforeSave == null || beforeSaveAsync == null),
        super(key: key);

  @override
  _DefaultUpdateScaffoldWidgetState<T> createState() =>
      _DefaultUpdateScaffoldWidgetState<T>();
}

class _DefaultUpdateScaffoldWidgetState<T extends UpdateableDatabaseItem<T>>
    extends State<DefaultUpdateScaffoldWidget<T>> {
  T originalItem;
  T copyOfUpdateableItem;
  bool isEditable;
  bool isNewItem;

  int validatePageIndex;
  Key mainKey;

  List<ScaffoldPageWidget<T>> pages;

  @override
  void initState() {
    super.initState();

    if (DefaultUpdateScaffoldWidget.databaseService == null)
      print(
          "[DefaultUpdateScaffoldWidget] Error: The static DatabseService was not set!");

    originalItem = widget.updateableItem;
    copyOfUpdateableItem = originalItem?.copy();
    isNewItem = originalItem.id == null;
    isEditable = isNewItem;

    if (isNewItem) widget.prepareNewItem?.call(context, copyOfUpdateableItem);

    validatePageIndex = null;
    mainKey = Key(Random().nextInt(10000).toString());
  }

  @override
  Widget build(BuildContext context) {
    pages ??= widget.updatePagesBuilder?.call(context);

    return Provider<T>.value(
      value: copyOfUpdateableItem,
      child: Provider<bool>.value(
        value: isEditable,
        child: DefaultScaffoldWidget(
          key: mainKey,
          title: widget.titleBuilder(copyOfUpdateableItem),
          showBackArrow: true,
          onWillPop: _onWillPop,
          menuButtons: _getMenuButtons(),
          updatePages: pages,
          child: widget.child,
          defaultPageIndex: validatePageIndex ?? 0,
        ),
      ),
    );
  }

  List<Widget> _getMenuButtons() {
    if (!widget.hasEditButton) return widget.menuButtons;

    if (isEditable)
      return <Widget>[
        SimpleScaffoldMenuButtonWidget(
          text: Text("Abbrechen"),
          onTap: () => _abortEditing(),
          tooltip: "Bearbeitung abbrechen",
        ),
        SimpleScaffoldMenuButtonWidget(
          text: Text("Speichern"),
          onTap: () => _validate(),
          tooltip: "Änderungen speichern",
        )
      ]..addAll(widget.editMenuButtons ?? widget.menuButtons);
    else
      return <Widget>[
        IconScaffoldMenuButtonWidget(
          icon: Icons.edit,
          text: Text("Bearbeiten"),
          onTap: () async {
            setState(() {
              isEditable = true;
            });
          },
          tooltip: "Eigenschaften bearbeiten",
        )
      ]..addAll(widget.menuButtons);
  }

  Future<bool> _onWillPop() async {
    if (isEditable) {
      if (copyOfUpdateableItem.equals(originalItem))
        return true;
      else {
        bool result = false;
        await showCustomDialog(
            context,
            SaveOrAbortDialog(
              onAbort: () {
                result = true;
              },
              onSave: () {
                result = false;
                _validate();
              },
            ));
        return result;
      }
    } else
      return true;
  }

  void _abortEditing() {
    widget.onAbort?.call(context, copyOfUpdateableItem, originalItem);
    setState(() {
      if (isNewItem) {
        Navigator.pop(context);
      } else {
        isEditable = false;
        copyOfUpdateableItem = originalItem?.copy();
      }
    });
  }

  bool _validate() {
    if (widget.validator != null) {
      if (widget.validator.isValid()) {
        _saveItem();
        return true;
      } else {
        setState(() {});
        return false;
      }
    }

    if (pages != null)
      for (var i = 0; i < pages.length; i++) {
        if (pages[i].isValid != null &&
            !pages[i].isValid(copyOfUpdateableItem)) {
          setState(() {
            validatePageIndex = i;
            mainKey = Key(Random().nextInt(10000).toString());
          });
          Future.delayed(
              Duration(milliseconds: 100), () => pages[i].validator?.isValid());
          return false;
        }
      }

    validatePageIndex = null;
    _saveItem();
    return true;
  }

  void _saveItem() async {
    if (widget.beforeSave != null)
      widget.beforeSave(context, copyOfUpdateableItem);

    if (widget.beforeSaveAsync != null)
      await widget.beforeSaveAsync(context, copyOfUpdateableItem);

    if (isNewItem) {
      var itemRef = await DefaultUpdateScaffoldWidget.databaseService
          .collectionService<T>()
          .add(copyOfUpdateableItem);
      widget.afterSave?.call(context, itemRef);
      Navigator.pop(context);
    } else {
      isEditable = false;
      DefaultUpdateScaffoldWidget.databaseService
          .collectionService<T>()
          .update(copyOfUpdateableItem);
      originalItem = copyOfUpdateableItem.copy();
    }

    setState(() {});
  }
}
